###########################################################
###                                                     ###
###   Some Sage code to verify the case enumerations    ###
###   in the proofs of the article                      ###
###   ``Substitutive Arnoux-Rauzy sequences             ###
###     have pure discrete spectrum''                   ###
###   by Valerie Berthe, Timo Jolivet, Anne Siegel      ###
###                                                     ###
###   Timo Jolivet, 28/08/2011                          ###
###                                                     ###
###   Works best with Sage 4.7.1 or higher.             ###
###   This is work in progress, use at your own risk!   ###
###                                                     ###
###########################################################

from sage.combinat.e_one_star import Face, Patch, E1Star

####################################################
####################################################
### Various sets of patterns used in the article ###
####################################################
####################################################

# Unit cube and single-face patterns
U = Patch([Face([0,0,0], 1), Face([0,0,0], 2), Face([0,0,0], 3)])

# Two-face edge-connected patterns
E11a = Patch([Face([0,0,0], 1), Face([0,1,0], 1)])
E11b = Patch([Face([0,0,0], 1), Face([0,0,1], 1)])
E22a = Patch([Face([0,0,0], 2), Face([1,0,0], 2)])
E22b = Patch([Face([0,0,0], 2), Face([0,0,1], 2)])
E33a = Patch([Face([0,0,0], 3), Face([1,0,0], 3)])
E33b = Patch([Face([0,0,0], 3), Face([0,1,0], 3)])

E12a = Patch([Face([0,0,0], 1), Face([0,0,0], 2)])
E12b = Patch([Face([0,0,0], 1), Face([-1,1,0], 2)])
E13a = Patch([Face([0,0,0], 1), Face([0,0,0], 3)])
E13b = Patch([Face([0,0,0], 1), Face([-1,0,1], 3)])
E23a = Patch([Face([0,0,0], 2), Face([0,0,0], 3)])
E23b = Patch([Face([0,0,0], 2), Face([0,-1,1], 3)])

LE = [E11a, E11b, E22a, E22b, E33a, E33b, \
      E12a, E12b, E13a, E13b, E23a, E23b]

# Two-face vertex-connected patterns
V11a = Patch([Face([0,0,0], 1), Face([0,1,1], 1)])
V11b = Patch([Face([0,0,0], 1), Face([0,1,-1], 1)])
V22a = Patch([Face([0,0,0], 2), Face([1,0,1], 2)])
V22b = Patch([Face([0,0,0], 2), Face([1,0,-1], 2)])
V33a = Patch([Face([0,0,0], 3), Face([1,1,0], 3)])
V33b = Patch([Face([0,0,0], 3), Face([1,-1,0], 3)])

V12a = Patch([Face([0,0,0], 1), Face([0,0,1], 2)])
V12b = Patch([Face([0,0,0], 2), Face([0,0,1], 1)])
V12c = Patch([Face([0,0,0], 1), Face([-1,0,1], 2)])
V12d = Patch([Face([0,0,0], 2), Face([0,-1,1], 1)])
V12e = Patch([Face([0,0,0], 1), Face([-1,1,1], 2)])
V12f = Patch([Face([0,0,0], 2), Face([1,-1,1], 1)])

V13a = Patch([Face([0,0,0], 1), Face([0,1,0], 3)])
V13b = Patch([Face([0,0,0], 3), Face([0,1,0], 1)])
V13c = Patch([Face([0,0,0], 1), Face([-1,1,0], 3)])
V13d = Patch([Face([0,0,0], 3), Face([0,1,-1], 1)])
V13e = Patch([Face([0,0,0], 1), Face([-1,1,1], 3)])
V13f = Patch([Face([0,0,0], 3), Face([1,1,-1], 1)])

V23a = Patch([Face([0,0,0], 2), Face([1,0,0], 3)])
V23b = Patch([Face([0,0,0], 3), Face([1,0,0], 2)])
V23c = Patch([Face([0,0,0], 2), Face([1,-1,0], 3)])
V23d = Patch([Face([0,0,0], 3), Face([1,0,-1], 2)])
V23e = Patch([Face([0,0,0], 2), Face([1,-1,1], 3)])
V23f = Patch([Face([0,0,0], 3), Face([1,1,-1], 2)])

LV = [V11a, V11b, V22a, V22b, V33a, V33b, \
      V12a, V12b, V12c, V12d, V12e, V12f, \
      V13a, V13b, V13c, V13d, V13e, V13f, \
      V23a, V23b, V23c, V23d, V23e, V23f]

# The above list LV without the patterns V11a, V22a, V33a
LVAR = list(LV)
LVAR.remove(V11a)
LVAR.remove(V22a)
LVAR.remove(V33a)

# Set of covering patterns LAR
E11aAR = Patch([Face([0,0,0], 3), Face([0,0,0], 1), Face([0,1,0], 1)])
E11bAR = Patch([Face([0,0,0], 2), Face([0,0,0], 1), Face([0,0,1], 1)])
E22aAR = Patch([Face([0,0,0], 3), Face([0,0,0], 2), Face([1,0,0], 2)])
E22bAR = Patch([Face([0,0,0], 1), Face([0,0,0], 2), Face([0,0,1], 2)])
E33aAR = Patch([Face([0,0,0], 2), Face([0,0,0], 3), Face([1,0,0], 3)])
E33bAR = Patch([Face([0,0,0], 1), Face([0,0,0], 3), Face([0,1,0], 3)])

LAR = [E12a, E12b, E13a, E13b, E23a, E23b, E11aAR, E11bAR, E22aAR, E22bAR, E33aAR, E33bAR]

# Arnoux-Rauzy dual substitutions
AR1 = E1Star(WordMorphism({1:[1], 2:[2,1], 3:[3,1]}))
AR2 = E1Star(WordMorphism({1:[1,2], 2:[2], 3:[3,2]}))
AR3 = E1Star(WordMorphism({1:[1,3], 2:[2,3], 3:[3]}))

####################################################
####################################################
### Tools to study preimages by AR substitutions ###
####################################################
####################################################

# The vectors below are used in the function is_connected_2
connected_2_11 = map(vector, [(0,1,0), (0,0,1), (0,1,-1), (0,1,1), (0,-1,0), (0,0,-1), (0,-1,1), (0,-1,-1)])
connected_2_22 = map(vector, [(1,0,0), (0,0,1), (1,0,-1), (1,0,1), (-1,0,0), (0,0,-1), (-1,0,1), (-1,0,-1)])
connected_2_33 = map(vector, [(1,0,0), (0,1,0), (1,-1,0), (1,1,0), (-1,0,0), (0,-1,0), (-1,1,0), (-1,-1,0)])
connected_2_12 = map(vector, [(0,0,0), (-1,1,0), (0,0,1), (0,0,-1), (0,1,-1), (-1,1,-1), (-1,1,1), (-1,0,1)])
connected_2_13 = map(vector, [(0,0,0), (-1,0,1), (0,1,0), (0,-1,0), (-1,1,0), (-1,1,1), (-1,-1,1), (0,-1,1)])
connected_2_23 = map(vector, [(0,0,0), (0,-1,1), (1,0,0), (-1,0,0), (-1,0,1), (-1,-1,1), (1,-1,1), (1,-1,0)])
connected_2_21 = [-v for v in connected_2_12]
connected_2_31 = [-v for v in connected_2_13]
connected_2_32 = [-v for v in connected_2_22]

def is_connected_2(P):
    r"""
    Tests whether the two-face pattern P is connected.
    """
    f1, f2 = list(P)
    t1 = f1.type()
    t2 = f2.type()
    v1 = f1.vector()
    v2 = f2.vector()

    if (t1,t2) == (1,1):
        return v2 - v1 in connected_2_11
    elif (t1,t2) == (2,2):
        return v2 - v1 in connected_2_22
    elif (t1,t2) == (3,3):
        return v2 - v1 in connected_2_33
    elif (t1,t2) == (1,2):
        return v2 - v1 in connected_2_12
    elif (t1,t2) == (1,3):
        return v2 - v1 in connected_2_13
    elif (t1,t2) == (2,3):
        return v2 - v1 in connected_2_23
    elif (t1,t2) == (2,1):
        return v2 - v1 in connected_2_21
    elif (t1,t2) == (3,1):
        return v2 - v1 in connected_2_31
    elif (t1,t2) == (3,2): 
        return v2 - v1 in connected_2_32 

def is_admissible_2(P):
    r"""
    Tests whether there exists a discrete plane that contains a translation of the two-face pattern P.
    Warning: this function is slow because it calls solve.
    """
    f1, f2 = list(P)
    x1, y1, z1 = f1.vector()
    x2, y2, z2 = f2.vector()
    t1 = f1.type()
    t2 = f2.type()
    a = var('a')
    b = var('b')
    c = var('c')
    if t1 == 1: t1 = a
    if t1 == 2: t1 = b
    if t1 == 3: t1 = c
    if t2 == 1: t2 = a
    if t2 == 2: t2 = b
    if t2 == 3: t2 = c

    eq1 = [-t1 < a*(x2-x1) + b*(y2-y1) + c*(z2-z1)]
    eq2 = [a*(x2-x1) + b*(y2-y1) + c*(z2-z1) < t2]
    eq3 = [a > 0, b > 0, c > 0]
    S = solve(eq1 + eq2 + eq3, a, b, c)

    return bool(S)

def preimage_face_AR(i, f):
    r"""
    Returns the preimages of the face f by AR_i (i = 1,2,3).
    """
    x, y, z = f.vector()
    t = f.type()
    L = []
    if i == 1:
        if t == 1: L += [Face([x+y+z, y, z], 1)]
        if t == 2: L += [Face([x+y+z, y, z], 1), Face([x+y+z-1, y, z], 2)]
        if t == 3: L += [Face([x+y+z, y, z], 1), Face([x+y+z-1, y, z], 3)]
    if i == 2:
        if t == 1: L += [Face([x, x+y+z, z], 2), Face([x, x+y+z-1, z], 1)]
        if t == 2: L += [Face([x, x+y+z, z], 2)]
        if t == 3: L += [Face([x, x+y+z, z], 2), Face([x, x+y+z-1, z], 3)]
    if i == 3:
        if t == 1: L += [Face([x, y, x+y+z], 3), Face([x, y, x+y+z-1], 1)]
        if t == 2: L += [Face([x, y, x+y+z], 3), Face([x, y, x+y+z-1], 2)]
        if t == 3: L += [Face([x, y, x+y+z], 3)]
    return L

def preimage_AR_2(i, P):
    r"""
    Returns the preimages by AR_i of the two-face pattern P = [f1, f2]
    that are of the form [g1, g2], such that:
        - f1 \in AR_i(g1),
        - f2 \in AR_i(g2),
        - a translation of [g1, g2] belongs to some discrete plane.
    This function is slow because it calls is_admissible_2,
    which is slow because it calls solve.
    """
    f1, f2 = list(P)
    L = []
    L1 = preimage_face_AR(i, f1)
    L2 = preimage_face_AR(i, f2)
    for g1 in L1:
        for g2 in L2:
            Q = Patch([g1, g2]) # we know that g1 != g2 because AR_i is injective and f1 != f2
            if is_admissible_2(Q):
                L.append(Q)
    return L

#################################################
#################################################
### Case enumerations performed in the proofs ###
#################################################
#################################################

def badcases_AR(i, L=LVAR):
    r"""
    Returns the list of triples (i,P,Q) such that for all P \in L,
    Q \subset AR_i(P) and Q is a disconnected pattern of two faces.
    (The patterns of L must be of two faces only.)
    """
    badcases = []
    for P in L:
        for Q in preimage_AR_2(i, P):
            if not is_connected_2(Q):
                badcases.append([i, P, Q])
                print i, P, Q
    return badcases

def print_badcases_AR(L, to_list=False, tikz=False):
    r"""
    A handy function to visualize the disconnected preimages.
    """
    if not tikz:
        Lg = []
        for x in L:
            x[2].repaint()
            G = Graphics()
            G += x[2].plot()
            G += AR[x[0]-1](x[2]).translate([-5,5,0]).plot()
            G += x[1].translate([0,0,4]).plot()
            G += text('AR_%d'%(x[0]), [2,5])
            G += text(x[2].__repr__()[8:-1], [1,-3])
            if to_list:
                Lg += G
            elif not to_list:
                G.show(axes=false)
        if to_list:
            return Lg
    elif tikz:
        s = ''
        s += '\\begin{tabular}{c|c}%\n'
        s += "$f \cup g$ & $f' \cup g'$ \\\\\n"
        for x in L:
            s += '\hline\n'
            x[1].repaint([(0.8,0.8,0.8)])
            x[2].repaint([(0.8,0.8,0.8)])
            s += '\myvcenter{%\n'
            s += x[2].plot_tikz()
            s += '}\n  &\n'
            s += '\myvcenter{%\n'
            s += x[1].plot_tikz()
            s += '}\n  \\\\\n'
        s += '\\end{tabular}'
        return s

def print_LVBAD(L):
    s = ''
    for x in L:
        x[1].repaint([(0.8,0.8,0.8)])
        s += '\hline\n'
        s += '%d &\n'%x[0]
        s += '\myvcenter{%\n'
        s += x[1].plot_tikz()
        s += '}\n  \\\\\n'
    return s

#LARBAD = badcases_AR(1) + badcases_AR(2) + badcases_AR(3) 

######################################################################
######################################################################
### The graph used to show that an annulus is always formed from U ###
######################################################################
######################################################################

# The set of faces at distance <= 1 from the origin. (Used in the function restrict_annulus.)
thickness_1 = set([ \
    Face([0,0,0],1), Face([0,0,0],2), Face([0,0,0],3), \
    Face([0,1,0],1), Face([-1,1,0],2), \
    Face([0,0,1],1), Face([-1,0,1],3), \
    Face([1,0,0],2), Face([1,-1,0],1), \
    Face([0,0,1],2), Face([0,-1,1],3), \
    Face([1,0,0],3), Face([1,0,-1],1), \
    Face([0,1,0],3), Face([0,1,-1],2), \
    Face([1,-1,-1],1), Face([1,0,-1],2), Face([1,-1,0],3), \
    Face([0,1,-1],1), Face([-1,1,-1],2), Face([-1,1,0],3), \
    Face([0,-1,1],1), Face([-1,0,1],2), Face([-1,-1,1],3), \
    Face([0,1,1],1), Face([-1,1,1],2), Face([-1,1,1],3), \
    Face([1,-1,1],1), Face([1,0,1],2), Face([1,-1,1],3), \
    Face([1,1,-1],1), Face([1,1,-1],2), Face([1,1,0],3)])

def restrict_annulus(P):
    r"""
    Restrict a pattern to its subpattern that contains only the faces at distance <= 1 from the origin.
    Also colors the patch to distinguish the unit cube.
    """
    P = Patch(set(P).intersection(thickness_1))
    P.repaint([(0.8,0.8,0.8)])
    for f in P.faces_of_vector([0,0,0]):
        f.color((0.35,0.35,0.35))
    return P

E1rotation = E1Star(WordMorphism({1:[2],2:[3],3:[1]})) # Rotates a patch by -pi/3

# Graph patterns
gAR_1 = restrict_annulus(AR1(U))
gAR_2 = restrict_annulus(AR2(U))
gAR_3 = restrict_annulus(AR3(U))

gAR_12 = restrict_annulus((AR2*AR1)(U))
gAR_13 = restrict_annulus((AR3*AR1)(U))
gAR_21 = restrict_annulus((AR1*AR2)(U))
gAR_23 = restrict_annulus((AR3*AR2)(U))
gAR_31 = restrict_annulus((AR1*AR3)(U))
gAR_32 = restrict_annulus((AR2*AR3)(U))

gAR_123 = restrict_annulus((AR3*AR2*AR1)(U))
gAR_132 = restrict_annulus((AR2*AR3*AR1)(U))
gAR_213 = E1rotation(gAR_132, 2)
gAR_231 = E1rotation(gAR_123, 2)
gAR_312 = E1rotation(gAR_123)
gAR_321 = E1rotation(gAR_132)

gAR_1231 = restrict_annulus((AR1*AR3*AR2*AR1)(U)) + Patch([Face([0,1,-1],3,(0.8,0.8,0.8))])
gAR_1321 = restrict_annulus((AR1*AR2*AR3*AR1)(U)) + Patch([Face([0,-1,1],2,(0.8,0.8,0.8))])
gAR_2132 = E1rotation(gAR_1321,2)
gAR_2312 = E1rotation(gAR_1231,2)
gAR_3123 = E1rotation(gAR_1231)
gAR_3213 = E1rotation(gAR_1321)

L_graph_AR = [gAR_1, gAR_2, gAR_3, gAR_12, gAR_13, gAR_21, gAR_23, gAR_31, gAR_32, gAR_123, gAR_132, gAR_213, gAR_231, gAR_312, gAR_321, gAR_1231, gAR_1321, gAR_2132, gAR_2312, gAR_3123, gAR_3213]



##### various stuff

# counterexample anneau strong cover (broken by AR2)
ARcounterex_P = U + [Face((1,0,0),2), Face((1,0,0),3)]
ARcounterex_P.repaint([(0.35,0.35,0.35)])
ARcounterex_A = Patch(gAR_1321)
ARcounterex_A -= U
ARcounterex_A -= Face([0,-1,1],2,(0.8,0.8,0.8))
ARcounterex_A = AR1(ARcounterex_A)
ARcounterex_A -= Face([0,-1,1],2,(0.8,0.8,0.8))

#### bad pattern P
P_vilain = Patch([Face((-1, 0, 2), 3), Face((0, 0, 1), 1), Face((0, 2, 1), 1), Face((0, 1, 0), 1), Face((-1, 0, 2), 1), Face((0, 0, 0), 2), Face((0, 0, 0), 3), Face((-1, 2, 2), 3), Face((-1, 1, 2), 3), Face((0, 0, 0), 1), Face((-1, 1, 2), 1)])

